@@ -1,10 +1,14 @@
-from typing import Dict, List, Set, Tuple
+from typing import Counter, Dict, Generator, List, Set, Tuple
 import collections
 import collections.abc
 import copy
+import csv
 import datetime
 import functools
+import io
 import itertools
+import os
+import pkg_resources
 import re
 import requests
 import urllib.request
@@ -16,6 +20,9 @@ from . import sets
 from . import utils
 
 
+LOCAL_CARDS = os.getenv("LOCAL_CARDS")
+
+
 class Card(utils.i18nMixin, utils.NamedMixin):
     #: official cards renaming not registered in cards "Aka" field
     _AKA = {
@@ -26,6 +33,9 @@ class Card(utils.i18nMixin, utils.NamedMixin):
         "Assamite": "Banu Haqim",
         "Follower of Set": "Ministry",
     }
+    _DISC_RENAMES = {
+        "Thaumaturgy": "Blood Sorcery",
+    }
     #: Actual ban dates
     _BAN_MAP = {
         "1995": datetime.date(1995, 11, 1),  # RTR 19951006
@@ -178,6 +188,10 @@ class Card(utils.i18nMixin, utils.NamedMixin):
             "PB": "Brujah antitribu",
             "PM": "Malkavian antitribu",
             "PTz": "Tzimisce",
+            "SKB": "Starter Kit Brujah antitribu",
+            "SKM": "Starter Kit Malkavian antitribu",
+            "SKTr": "Starter Kit Tremere antitribu",
+            "SKTz": "Starter Kit Tzimisce",
         },
         "Tenth": {"A": "Tin A", "B": "Tin B"},
         "Anthology": {"LARP": "EC Berlin Edition"},
@@ -196,6 +210,13 @@ class Card(utils.i18nMixin, utils.NamedMixin):
             "PBh": "Banu Haqim",
             "PG": "Gangrel",
         },
+        "NB": {
+            "PM": "Malkavian",
+            "PN": "Nosferatu",
+            "PTo": "Toreador",
+            "PTr": "Tremere",
+            "PV": "Ventrue",
+        },
     }
     _REPRINTS_RELEASE_DATE = {
         ("KoT", "Reprint Bundle 1"): datetime.date(2018, 5, 5),
@@ -247,6 +268,8 @@ class Card(utils.i18nMixin, utils.NamedMixin):
         self.combo = None
         self.multidisc = None
         self.card_text = ""
+        # original VEKN value: use `sets` instead for more accessible info
+        self._set = ""
         self.sets = {}
         self.scans = {}
         self.banned = None
@@ -261,13 +284,15 @@ class Card(utils.i18nMixin, utils.NamedMixin):
         self.flavor_text = None
         self.draft = None
         # enriched properties (not directly in original CSV, but convenient)
+        self.ordered_sets = []  # sets in release order
         self.has_advanced = None  # same vampire appears as advanced in the same group
         self.has_evolution = None  # same vampire appears in a higher group
         self.is_evolution = None  # same vampire appears in a lower group
+        self.variants = {}  # variants of the same vampire (base, adv, evolution)
         self.name_variants = []  # variations you want to match when parsing a decklist
         self.rulings = {"text": [], "links": {}}
 
-    def diff(self, rhs):
+    def diff(self, rhs) -> Dict[str, Tuple[str, str]]:
         res = {}
         if self.name != rhs.name:
             res["name"] = [self.name, rhs.name]
@@ -355,6 +380,18 @@ class Card(utils.i18nMixin, utils.NamedMixin):
             ret = "The " + ret[:-5]
         return ret.replace("(TM)", "™")
 
+    @property
+    @functools.lru_cache(1)
+    def _key(self) -> str:
+        """Used internally for advanced / evolutions / variants computations"""
+        if self.group == "ANY":
+            key = "ANY"
+        else:
+            key = f"G{self.group}"
+        if self.adv:
+            key += " ADV"
+        return key
+
     def get_suffix(self, minimal=False) -> str:
         suffixes = []
         if self.group and (self.is_evolution or not minimal):
@@ -384,6 +421,7 @@ class Card(utils.i18nMixin, utils.NamedMixin):
             {
                 k: v
                 for k, v in list(self.__dict__.items())
+                # add usual names for convenience
                 + [("name", self.name), ("printed_name", self.printed_name)]
                 if k not in ["crypt", "library", "vekn_name"]
             }
@@ -391,6 +429,7 @@ class Card(utils.i18nMixin, utils.NamedMixin):
 
     def from_json(self, state: Dict) -> None:
         """Get the card form a dict."""
+        # remove convenience names to avoid overriding the properties
         state.pop("name", None)
         state.pop("printed_name", None)
         self.__dict__.update(state)
@@ -422,6 +461,7 @@ class Card(utils.i18nMixin, utils.NamedMixin):
 
         self.id = int(data["Id"])
         self._name = data["Name"]
+        self._set = data["Set"]
         self.aka = split("Aka", ";")
         self.types = split("Type", "/")
         self.clans = split("Clan", "/")
@@ -450,17 +490,25 @@ class Card(utils.i18nMixin, utils.NamedMixin):
         self.card_text = (
             data["Card Text"].replace("(D)", "Ⓓ").replace("{", "").replace("}", "")
         )
+        if "{" in data["Card Text"]:
+            self.text_change = True
+        else:
+            self.text_change = False
         for old_name, new_name in self._CLAN_RENAMES.items():
             self.card_text = self.card_text.replace(old_name, new_name)
+        for old_name, new_name in self._DISC_RENAMES.items():
+            self.card_text = self.card_text.replace(old_name, new_name)
         self.banned = (
             self._BAN_MAP[data["Banned"]].isoformat() if data["Banned"] else None
         )
-        self.artists = sorted(  # some cards have duplicated artists, eg. Ashur Tablets
-            set(
+        # remove potential duplicated artists (eg. Ashur Tablets)
+        # collections.Counter to keep the order (ordered dict with convenient init)
+        self.artists = list(
+            collections.Counter(
                 self._ARTISTS_FIXES.get(s, s)
                 for s in map(str.strip, re.split(r"[;,&]+(?!\sJr\.)", data["Artist"]))
                 if s
-            )
+            ).keys()
         )
         self.adv = bool_or_none("Adv")
         # group can be "any"
@@ -474,6 +522,7 @@ class Card(utils.i18nMixin, utils.NamedMixin):
         self.burn_option = bool_or_none("Burn Option")
         self.flavor_text = data["Flavor Text"] if "Flavor Text" in data else None
         self.draft = data["Draft"] if "Draft" in data else None
+
         # computations last: some properties (ie. name) are cached,
         # only use them once everything else is set
         self.sets = dict(
@@ -486,7 +535,12 @@ class Card(utils.i18nMixin, utils.NamedMixin):
             )
             if rarity
         )
-        if not self.sets:
+        if self.sets:
+            self.ordered_sets = sorted(
+                [s for s in self.sets if set_dict[s].release_date],
+                key=lambda x: set_dict[x].release_date,
+            )
+        else:
             warnings.warn(f"no set found for {self}")
         self.scans = {
             name: self._compute_url(
@@ -509,7 +563,7 @@ class Card(utils.i18nMixin, utils.NamedMixin):
         }
         self.url = self._compute_url()
 
-    def _compute_url(self, lang: str = None, expansion: str = None):
+    def _compute_url(self, lang: str = None, expansion: str = None) -> str:
         """Compute image URL for given language."""
         return (
             config.KRCG_STATIC_SERVER
@@ -520,7 +574,7 @@ class Card(utils.i18nMixin, utils.NamedMixin):
             + ".jpg"
         )
 
-    def _compute_legacy_url(self, lang: str = None, expansion: str = None):
+    def _compute_legacy_url(self, lang: str = None, expansion: str = None) -> str:
         """Compute legacy image URL for given language."""
         return (
             config.KRCG_STATIC_SERVER
@@ -651,7 +705,7 @@ class CardMap(utils.FuzzyDict):
         """Use config.ALIASES as aliases"""
         super().__init__(aliases=config.ALIASES if aliases is None else aliases)
 
-    def __iter__(self):
+    def __iter__(self) -> Generator[Card, None, None]:
         """When iterating only, return each card once."""
         for key, card in self.items():
             if isinstance(key, int):
@@ -685,11 +739,33 @@ class CardMap(utils.FuzzyDict):
             for name in Card._AKA.get(card.id, []):
                 self[name] = self[card.id]
 
-    def load_from_vekn(self):
+    def load_from_vekn(self) -> None:
         """Load from official VEKN CSV files."""
         set_dict = sets.SetMap()
         # download the zip files containing the official CSV
-        main_files = utils.get_zip_csv(self._VEKN_CSV[0], *self._VEKN_CSV[1])
+        if LOCAL_CARDS:
+            main_files = [
+                csv.DictReader(
+                    io.TextIOWrapper(
+                        pkg_resources.resource_stream("cards", "vtessets.csv"),
+                        encoding="utf-8-sig",
+                    )
+                ),
+                csv.DictReader(
+                    io.TextIOWrapper(
+                        pkg_resources.resource_stream("cards", "vtescrypt.csv"),
+                        encoding="utf-8-sig",
+                    )
+                ),
+                csv.DictReader(
+                    io.TextIOWrapper(
+                        pkg_resources.resource_stream("cards", "vteslib.csv"),
+                        encoding="utf-8-sig",
+                    )
+                ),
+            ]
+        else:
+            main_files = utils.get_zip_csv(self._VEKN_CSV[0], *self._VEKN_CSV[1])
         i18n_files = {
             lang: utils.get_zip_csv(url, *filenames)
             for lang, (url, filenames) in self._VEKN_CSV_I18N.items()
@@ -736,7 +812,7 @@ class CardMap(utils.FuzzyDict):
         # all name variants computed, now we can map all those name in the dict
         self._map_names()
 
-    def _set_enriched_properties(self):
+    def _set_enriched_properties(self) -> None:
         """Set enriched properties on cards.
 
         This method sets addition information related to the cards
@@ -745,6 +821,7 @@ class CardMap(utils.FuzzyDict):
         - card.has_advanced
         - card.has_evolution
         - card.is_evolution
+        - card.variants
         """
         # first compute, for cards with same name (crypt cards only),
         # which were the one to appear first (first_group)
@@ -754,9 +831,12 @@ class CardMap(utils.FuzzyDict):
         same_name = {k: v for k, v in same_name.items() if len(v) > 1}
         for name, cards in same_name.items():
             groups = collections.defaultdict(list)
+            variants = {}
             for card in cards:
                 groups[card.group].append(card)
+                variants[card._key] = card.id
             groups = sorted(groups.items(), key=lambda a: a[0])
+
             for i, (_group, cards) in enumerate(groups):
                 if len(cards) > 1:
                     assert sum(bool(c.adv) for c in cards) == 1, "bad advanced mark"
@@ -767,6 +847,10 @@ class CardMap(utils.FuzzyDict):
                         card.has_evolution = True
                     if i > 0:
                         card.is_evolution = True
+                    card.variants = {
+                        k: v for k, v in variants.items() if k != card._key
+                    }
+
         # now compute variants - cards in first_group can omit the group suffix
         # advanced version can never omit the suffix
         for card in self:
@@ -782,7 +866,7 @@ class CardMap(utils.FuzzyDict):
                 name_variants = name_variants[1:]  # first name is the actual name
                 card.i18n_set(lang, {"name_variants": name_variants})
 
-    def _variants(self, name, card):
+    def _variants(self, name, card) -> Generator[str, None, None]:
         suffix = card.get_suffix()
         if suffix:
             suffix = f" ({suffix})"
@@ -795,7 +879,7 @@ class CardMap(utils.FuzzyDict):
             else:
                 yield from self._word_variants(name, "")
 
-    def _word_variants(self, name, suffix):
+    def _word_variants(self, name, suffix) -> Generator[str, None, None]:
         if "(TM)" in name:
             yield from self._word_variants(name.replace("(TM)", "™"), suffix)
         if name[-5:] == ", The":
@@ -804,7 +888,7 @@ class CardMap(utils.FuzzyDict):
         if name[:4] == "The ":
             yield from self._comma_splits(name[4:] + ", The", suffix)
 
-    def _comma_splits(self, name, suffix):
+    def _comma_splits(self, name, suffix) -> Generator[str, None, None]:
         while True:
             yield name + suffix
             name = name.rsplit(",", 1)
@@ -823,12 +907,23 @@ class CardMap(utils.FuzzyDict):
                 self[cid].rulings["text"].append(ruling.text)
                 for ref, link in ruling.links.items():
                     self[cid].rulings["links"][ref] = link
+            for card_reference in re.findall(r"{[^}]+}", ruling.text):
+                card_reference = card_reference[1:-1]
+                if card_reference not in self:
+                    warnings.warn(
+                        f"Rulings: {cid}|{name} mentions unknown card {card_reference}"
+                    )
+                if self[card_reference].usual_name != card_reference:
+                    warnings.warn(
+                        f"Rulings: {cid}|{name} mentions {card_reference} "
+                        f"instead of '{self[card_reference].name}'"
+                    )
 
     def to_json(self) -> Dict:
         """Return a compact list representation for JSON serialization."""
         return [card.to_json() for card in self]
 
-    def from_json(self, state: Dict):
+    def from_json(self, state: Dict) -> None:
         """Initialize from a JSON list."""
         for dict_ in state:
             card = Card()
@@ -858,7 +953,7 @@ class CardTrie:
         for lang, trans in card.i18n_variants(self.attribute):
             self.tries[lang[:2]].add(trans, card)
 
-    def search(self, text: str, lang: str = None) -> Dict[str, collections.Counter]:
+    def search(self, text: str, lang: str = None) -> Dict[str, Counter[Card]]:
         """Search for `text`, in english and optional `lang`.
 
         Returns:
@@ -971,10 +1066,10 @@ class CardSearch:
         self._set_dimensions_enums = None
         self._normalized_set_enum_map = None
 
-    def __bool__(self):
+    def __bool__(self) -> bool:
         return bool(self._all)
 
-    def clear(self):
+    def clear(self) -> None:
         """Clear content."""
         for attr in self.set_dimensions + self.trie_dimensions:
             try:
@@ -982,7 +1077,7 @@ class CardSearch:
             except TypeError:
                 pass
 
-    def add(self, card: Card):
+    def add(self, card: Card) -> None:
         """Add a card to the engine."""
         self._set_dimensions_enums = None
         self._normalized_set_enum_map = None
@@ -1037,11 +1132,11 @@ class CardSearch:
         self._handle_exceptions(card)
 
     @property
-    def dimensions(self):
+    def dimensions(self) -> List[str]:
         return self.trie_dimensions + self.set_dimensions + ["text"]
 
     @property
-    def set_dimensions_enums(self):
+    def set_dimensions_enums(self) -> Dict[str, List[str]]:
         if not self._set_dimensions_enums:
             self._set_dimensions_enums = {
                 attr: sorted(set(getattr(self, attr).keys()))
@@ -1112,7 +1207,7 @@ class CardSearch:
         return result
 
     @property
-    def _normalized_map(self):
+    def _normalized_map(self) -> Dict[str, Dict[str, str]]:
         """Used by __call__ (the search itself) to match set dimensions values
 
         This allows to match values case insensitively
@@ -1128,7 +1223,7 @@ class CardSearch:
             }
         return self._normalized_set_enum_map
 
-    def _handle_text(self, card):
+    def _handle_text(self, card) -> None:
         """Helper handling card text."""
         self.card_text.add(card)
         if card.flavor_text:
@@ -1136,7 +1231,7 @@ class CardSearch:
         if card.draft:
             self.bonus["draft"].add(card.draft, card)
 
-    def _handle_disciplines(self, card):
+    def _handle_disciplines(self, card) -> None:
         """Helper handling card disciplines."""
         for discipline in card.disciplines:
             self.discipline[discipline].add(card)
@@ -1154,7 +1249,7 @@ class CardSearch:
         if card.multidisc:
             self.discipline["choice"].add(card)
 
-    def _handle_sect(self, card):
+    def _handle_sect(self, card) -> None:
         """Helper handling sects."""
         if card.crypt:
             if re.search(r"^(\[MERGED\] )?Sabbat", card.card_text):
@@ -1182,7 +1277,7 @@ class CardSearch:
             if re.search(r"Requires a( ready|n) (I|i)ndependent", card.card_text):
                 self.sect["Independent"].add(card)
 
-    def _handle_stealth_intercept(self, card):
+    def _handle_stealth_intercept(self, card) -> None:
         """Helper handling stealth and intercept."""
         if re.search(r"\+(\d|X)\s+(i|I)ntercept", card.card_text):
             self.bonus["Intercept"].add(card)
@@ -1206,7 +1301,7 @@ class CardSearch:
         if re.search(r"attempt fails", card.card_text):
             self.bonus["Stealth"].add(card)
 
-    def _handle_titles(self, card):
+    def _handle_titles(self, card) -> None:
         """Helper handling titles, votes and city."""
         if card.title:
             self.title[card.title].add(card)
@@ -1250,7 +1345,7 @@ class CardSearch:
                 self.title["Magaji"].add(card)
                 self.sect["Laibon"].add(card)
 
-    def _handle_exceptions(self, card):
+    def _handle_exceptions(self, card) -> None:
         """Search exceptions not handled automatically."""
         # The Baron has his name in card text, but is not an Anarch Baron.
         if card.id == 200167:  # "The Baron"
